function [UPPOUT,LOWOUT,JETOUT,SIGOUT,VENOUT] = flute(LENGTH,VENT,SAMPLES,PRESSURE)

%
% [UPPOUT,LOWOUT,JETOUT,SIGOUT,VENOUT] = flute(LENGTH,VENT,SAMPLES)
% [UPPOUT,LOWOUT,JETOUT,SIGOUT,VENOUT] = flute(LENGTH,VENT,SAMPLES,PRESSURE)
%
% Arguments:
%
% LENGTH:   Flute length (delay line steps)
% VENT:     Vent location (expressed in delay line steps)
% SAMPLES:  Number of samples to be computed
% PRESSURE: Vector has that 'SAMPLES' random values in range [-1,1]
%
% Outputs:
%
% UPPOUT:   Output from the end of the upper delay line
% LOWOUT:   Output from the end of the lower delay line
% JETOUT:   Output from the end of the jet line
% SIGOUT:   Output from the open end of the flute
% VENOUT:   Output from the vent
%
% Use the command "wavwrite(SIGNAL/max(abs(SIGNAL))*127+128,44100,'filename')"
% to write the 'SIGNAL' vector as a wav-file.
%
% Flute model and the original implementation: Vesa Valimaki 1993
% This rewritten and de-bugged version:        Rami Hanninen 1995
%

% Constants:
%

FREQUENCY = 44100.0;          % Sampling frequency (Hz)
INTERVAL  = 1/FREQUENCY;      % Sampling interval (s)
LENGTH1   = LENGTH - 1;
STARTTIME = clock;

%
% Filter parameters:
%

RB0 = -0.4794;   % Reflection filter coefficients
RB1 = -0.2603;
DCA =  0.9950;   % DC killer (high-pass filter) coefficient

%
% Fractional Delay 3-Port (Finger hole model):
%
% (Implement parametrized interpolator parameter computation later)
%

INTERPOLATORDEGREE = 0;
INTERPOLATORCOEFFS = [ 1 ];

AR  = -0.8632; % Finger hole filter parameters
BR  = -(1+AR);
ARI = 0;
BRI = 0;

%
% Compute vent operation times
%

OPENVENTHERE  = round(1 * SAMPLES / 3);
CLOSEVENTHERE = round(2 * SAMPLES / 3);

%
% Variable Parameters of the Flute Model:
%

LOSSFREQ   =  0.6;          % Coeff for the freq-dependent loss filter
LOSSSCALE  =  1 - LOSSFREQ; % Use this to scale DC gain of the loss filter
LOSSGAIN   =  0.94;         % Freq-independent (DC) loss gain (0.92)
BACKGAIN   = -0.97;         % Real reflection coefficient for the mouth end
NOISEGAIN  =  4.0;          % Gain of the white noise signal
INPUTGAIN  =  100.0;        % Maximum amplitude of input
VOICEDGAIN =  0.02;         % Gain of the feedback from the tube to the nonlin
DIRFDBACK  = -0.1;          % Gain from the nonlinearity back to itself
SIGMIN     =  0.001;        % Input gain of the sigmoid function
SIGMOFFSET =  0.0;          % Offset of the sigmoid function
SIGMOUT    =  300.0;        % Output gain of the sigmoid function

% Coefficient of the leaky integrator (if 0.0, integrator is removed)

INTEGRA    =  0.0;

% Length of the jet line

JETLENGTH  = LENGTH;
JETLENGTH1 = JETLENGTH - 1;

%
% Delay lines and pointers:
%
% Jet:   --> 0123...N-1 -->
% Upper: --> 0123...N-1 -->
% Lower: <-- 0123...N-1 <--
%
% Delay line pointer point always to the newest (first) sample on the line.
%
% To get the newest sample from JET or UPPER, use PTR+1
% To get the oldest sample from JET or UPPER, use rem(PTR+LENGTH1,LENGTH)+1
% To get the newest sample from LOWER,        use PTR+1
% To get the oldest sample from LOWER,        use rem(PTR+1,LENGTH)+1
%
% NOTE: When pointer moves left, the data in delay line moves right
% NOTE: +1 must be added to all array indeces, because in Matlab table
%       indexes start from 1, not 0


JET      = zeros(1,JETLENGTH); % Delay line for the air jet
JETPTR   = 0;                  % Pointer to the jet delay line
UPPER    = zeros(1,LENGTH);    % Upper trail of the digital waveguide
UPPERPTR = LENGTH1;            % Pointer to the upper trail
LOWER    = zeros(1,LENGTH);    % Lower trail of the digital waveguide
LOWERPTR = 0;                  % Pointer to the lower trail

%
% Allocate memory for output vectors
%

UPPOUT = zeros(1,SAMPLES);
LOWOUT = zeros(1,SAMPLES);
JETOUT = zeros(1,SAMPLES);     
SIGOUT = zeros(1,SAMPLES);
VENOUT = zeros(1,SAMPLES);

%
% Input of the flute model
%

if exist('PRESSURE') == 0
  PRESSURE  = NOISEGAIN * 2 *(rand(1,SAMPLES) - 0.5);  % Input to the model
end

INPUTAMPL = INPUTGAIN*[linspace(0,1,2000) ones(1,SAMPLES-4000) zeros(1,2000)];

%PRESSURE = [-1000 zeros(1,SAMPLES-1)];  % For measuring the impulse response
%INPUTGAIN = [1 zeros(1,SAMPLES-1)];

%
% State variables
%

LOSSDELAY    = 0.0;  % Unit delay of the freq-dependent loss filter
REFLDELAYX   = 0.0;  % First unit delay of the reflection function
REFLDELAYY   = 0.0;  % Second unit delay of the reflection function
SIGMINPUT    = 0.0;  % Input of the sigmoid function
SIGMOUTPUT0  = 0.0;  % Output of the sigmoid function
SIGMOUTPUT1  = 0.0;  % Last output of the sigmoid function
DCXOUTPUT0   = 0.0;  % output of the DC killer filter
DCXOUTPUT1   = 0.0;  % Last output of the DC killer filter
INTEGRINPUT  = 0.0;  % Input of the leaky integrator
INTEGROUTPUT = 0.0;  % Output of the integrator
VENTOUTPUT   = 0.0;  % 3-port vent output

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% The computation of the flute model is described below.
% The parts of the waveguide flute algorithm are as follows
% (in this order):
%
%  - add input (noise signal) and signal from the integrator
%  - multiply them by the INPUTGAIN
%  - delay the above result using jet delay line
%  - feed the delayed signal into the sigmoid waveshaper
%  - add the output of the sigmoid function and the reflecting signal
%    from the lower delay line of the waveguide
%  - FRACTIONAL DELAY 3-PORT HERE !!!
%  - update the pointers to the delay lines (circular buffers) and move
%    them to the other end of the lines
%  - take a sample from the end of the upper trail and feed it to
%    the reflection filter
%  - run the reflection filter one step
%  - subtract the output of the reflection function from its input
%  - this result is the output of the model which has to be differentiated
%  - feed the output of the reflection filter to the lower delay line
%  - take a sample out of the other end (i.e., beginning) of the lower line
%  - add it to the output of the sigmoid function (multiplied by DIRFDBACK)
%  - multiply this result by VOICEDGAIN and feed it to the leaky integrator
%  - run one step of the integration (DC gain must be = 1)
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

for I = 1:SAMPLES

  % Print progress indicator

  if rem(I,100) == 0
    fprintf(1,'Sample %d (%.2f %%)\n',I,100*I/SAMPLES);
  end;

  % Open and close the vent

  if I == OPENVENTHERE
    JETLENGTH  = VENT;
    JETLENGTH1 = JETLENGTH - 1;
    BRI = BR;
  elseif I == CLOSEVENTHERE
    JETLENGTH  = LENGTH;
    JETLENGTH1 = JETLENGTH - 1;
    BRI = 0.0;
  end

  % Feed jet line

  JET(JETPTR+1) = INPUTAMPL(I) * (PRESSURE(I) + INTEGROUTPUT);

  % Sigmoid nonlinearity:

  SIGMINPUT   = JET(rem(JETPTR+JETLENGTH1,JETLENGTH)+1); % Get oldest sample
  SIGMOUTPUT1 = SIGMOUTPUT0;
  SIGMOUTPUT0 = SIGMOUT * tanh(INPUTAMPL(I) * SIGMOFFSET - SIGMIN * SIGMINPUT);

  % DC killer (a 1st order high-pass filter, direct form II)

  TEMP       = SIGMOUTPUT0 + DCA * DCXOUTPUT1;
  DCXOUTPUT0 = TEMP - DCXOUTPUT1;  % differentiation removes DC component
  DCXOUTPUT1 = TEMP;

  % Add (subtract) the output of the sigmoid function (DC killed) and the
  % reflected signal from the end of the lower delay line and feed the
  % result into the upper delay line.

  UPPER(UPPERPTR+1)=DCXOUTPUT0+BACKGAIN*LOWER(rem(LOWERPTR+1,LENGTH)+1);

  % FRACTIONAL DELAY 3-PORT:
  % ILINE = delay line of the interpolator
  % INTERPOLATORCOEFFS = Lagrange interpolator coefficients
  % DEILINE = delay line (state) of the deinterpolator
  % br & ar = coefficients of the reflection filter

  % NOTE: Upper and lower delay line pointers move into different directions.
  % NOTE: When for example the upper delay line pointer moves to the negative
  %       direction, it models that data in the delay line actually moves to
  %       the positive direction.
  %
  % Here we want to tap the upper and lower delay lines from the same spot,
  % measured from the beginning of the upper delay line. To get the
  % data element from the upper delay line that has been delayed N steps,
  % we just add N to the current upper pointer value and apply the remainder
  % operator. To get the corresponding data sample from the lower delay line
  % that flows into the opposite direction, we must add +1 to N because
  % what we actually want from the lower delay line is the sample that
  % has been delayed ((LENGTH-1)-N) steps. This equals to N'=(LENGTH-(N+1)).
  % Because the lower line flows to the negative direction, the actual
  % sample index is be calculated by substracting N' from the current
  % delay line pointer: (LOWERPTR-(LENGTH-(N+1))). This equals to
  % (LOWERPTR-LENGTH+N+1), and because of the modulo arithmetic, we can
  % drop the -LENGTH, so we get the final form: (LOWERPTR + N + 1).
  %
  % Because the delay lines flow into different directions, it is now easy to
  % get M subsequent data samples for the fractional delay interpolator.
  %
  %                          ::
  %                 --> HGFEDCBA**  -->
  %                <--  **ABCDEFGH <--
  %                          ::

  % Interpolate fraction delay (Can't use matlab vector notation because
  % of the required remainder operator ==> Must use a slow for-loop)

  for J = 0:INTERPOLATORDEGREE
    ILINE(J + 1) = UPPER(rem(UPPERPTR + VENT + J    ,LENGTH) + 1) + ...
                   LOWER(rem(LOWERPTR + VENT + J + 1,LENGTH) + 1);
  end;

  % 3-port computation

  VENTINPUT  = INTERPOLATORCOEFFS * ILINE';
  VENTOUTPUT = BRI * VENTINPUT - AR * VENTOUTPUT;

  % Deinterpolate fraction delay

  DEILINE = VENTOUTPUT * INTERPOLATORCOEFFS;

  for J = 0:INTERPOLATORDEGREE

    DUPPERPTR = rem(UPPERPTR + VENT + J    ,LENGTH);
    DLOWERPTR = rem(LOWERPTR + VENT + J + 1,LENGTH);

    UPPER(DUPPERPTR + 1) = UPPER(DUPPERPTR + 1) + DEILINE(J + 1);
    LOWER(DLOWERPTR + 1) = LOWER(DLOWERPTR + 1) + DEILINE(J + 1);

  end;

  % Boundary losses (1st order all-pole low-pass filter):

  LOSSDELAY = LOSSSCALE*LOSSGAIN*UPPER(rem(UPPERPTR+LENGTH1,LENGTH)+1) + ...
              LOSSFREQ*LOSSDELAY;

  REFLDELAYY = RB0 * LOSSDELAY + RB1 * (REFLDELAYX - REFLDELAYY);
  REFLDELAYX = LOSSDELAY;  % Update unit delay of the reflection filter

  % Feed reflected signal into the beginning of the lower delay line

  LOWER(LOWERPTR + 1) = REFLDELAYY;

  % Feedback loop:

  INTEGRINPUT = VOICEDGAIN * (LOWER(rem(LOWERPTR + 1,LENGTH) + 1) + ...
                              DIRFDBACK * DCXOUTPUT0);

  % Integration:

  INTEGROUTPUT = (1 - INTEGRA) * INTEGRINPUT + INTEGRA * INTEGROUTPUT;

  % Output of the model:
  % Subtract output of refl. filter from its input (and differentiate):

  UPPOUT(1,I) = UPPER(rem(UPPERPTR + LENGTH1,LENGTH) + 1);
  LOWOUT(1,I) = LOWER(rem(LOWERPTR + 1,LENGTH) + 1);
  JETOUT(1,I) = JET(rem(JETPTR + JETLENGTH1,JETLENGTH) + 1);
  SIGOUT(1,I) = LOSSDELAY - REFLDELAYY;
  VENOUT(1,I) = VENTOUTPUT;

  %SIGOUT(1,I) = SIGMOUTPUT0;
  %DCXOUT(1,I) = DCXOUTPUT0;

  % Move delay line pointers

  JETPTR   = rem(JETPTR   + JETLENGTH1,JETLENGTH); % -->
  UPPERPTR = rem(UPPERPTR + LENGTH1,   LENGTH);    % -->
  LOWERPTR = rem(LOWERPTR + 1,         LENGTH);    % <--

end; % End of the main loop

READYTIME = clock;
EXETIME   = etime(READYTIME,STARTTIME)
