function [UPPOUT,LOWOUT,JETOUT,SIGOUT,VENOUT] = flute(LENGTH,VENTS,DEGREE,SAMPLES,PRESSURE)

% [UPPOUT,LOWOUT,JETOUT,SIGOUT,VENOUT] = flute(L,VENTS,DEGREE,SAMPLES)
% [UPPOUT,LOWOUT,JETOUT,SIGOUT,VENOUT] = flute(L,VENTS,DEGREE,SAMPLES,PRESSURE)
%
% Arguments:
%
% L:        Flute and jet line lengths (FLU or [FLU JET1 JET2])
% VENTS:    Vent locations ([V1 V2 ... VN])
% DEGREE:   Fractional delay interpolator degree
% SAMPLES:  Number of samples to be computed
% PRESSURE: Vector that has 'SAMPLES' random values in range [-1,1]
%
% FLU:      Flute length (upper and lower delay line lengths)
% JET1:     Jet line length when vents are closed (default: FLU)
% JET2:     Jet line length when vents are open   (default: min(VENTS))
%
% Lengths and positions are expressed in unit delay steps. Vent locations
% may also contain a fractional part.
%
% Outputs:
%
% UPPOUT:   Output from the end of the upper delay line (  1   x SAMPLES)
% LOWOUT:   Output from the end of the lower delay line (  1   x SAMPLES)
% JETOUT:   Output from the end of the jet line         (  1   x SAMPLES)
% SIGOUT:   Output from the open end of the flute       (  1   x SAMPLES)
% VENOUT:   Output from the vents                       (VENTS x SAMPLES)
%
% Use the command "wavwrite(SIGNAL/max(abs(SIGNAL))*127+128,44100,'filename')"
% to write the 'SIGNAL' vector as a wav-file.
%
% Flute model and the original implementation:    Vesa Valimaki 1993
% This completely rewritten and expanded version: Rami Hanninen 1995

%------------------------------------------------------------------------------
% Resolve delay line lengths
%

FLUTELENGTH = LENGTH(1);

if size(LENGTH,2) == 3
  JETLENGTH1 = LENGTH(2);
  JETLENGTH2 = LENGTH(3);
else
  JETLENGTH1 = FLUTELENGTH;
  JETLENGTH2 = round(min(VENTS));
end

JETLENGTH  = FLUTELENGTH;

%------------------------------------------------------------------------------
% Constants
%

FREQUENCY = 44100.0;          % Sampling frequency (Hz)
INTERVAL  = 1/FREQUENCY;      % Sampling interval (s)
STARTTIME = clock;

%------------------------------------------------------------------------------
% Filter parameters
%

RB0 = -0.4794;   % Reflection filter coefficients
RB1 = -0.2603;
DCA =  0.9950;   % DC killer (high-pass filter) coefficient

%------------------------------------------------------------------------------
% Fractional Delay 3-Port (Finger hole model):
%
% Note that because COEFFS will contain a large number of zero elements,
% it should be realised as a sparse matrix.
%

DEGREE    = 2;
NROFVENTS = size(VENTS,2);
COEFFS    = sparse(FLUTELENGTH,NROFVENTS); % sparse() <==> zeros()

for I = 1:NROFVENTS
  COEFFS(:,I) = fluteinterp(FLUTELENGTH,DEGREE,VENTS(I))';
end

%------------------------------------------------------------------------------
% Finger hole filter parameters
%

AR  = -0.8632;
BR  = -(1+AR);
ARI = 0;
BRI = 0;

%------------------------------------------------------------------------------
% Compute vent operation times
%

OPENVENTHERE  = round(1 * SAMPLES / 3);
CLOSEVENTHERE = round(2 * SAMPLES / 3);

%------------------------------------------------------------------------------
% Variable parameters of the Flute Model:
%

LOSSFREQ   =  0.6;          % Coeff for the freq-dependent loss filter
LOSSSCALE  =  1 - LOSSFREQ; % Use this to scale DC gain of the loss filter
LOSSGAIN   =  0.94;         % Freq-independent (DC) loss gain (0.92)
BACKGAIN   = -0.97;         % Real reflection coefficient for the mouth end
NOISEGAIN  =  4.0;          % Gain of the white noise signal
INPUTGAIN  =  100.0;        % Maximum amplitude of input
VOICEDGAIN =  0.02;         % Gain of the feedback from the tube to the nonlin
DIRFDBACK  = -0.1;          % Gain from the nonlinearity back to itself
SIGMIN     =  0.001;        % Input gain of the sigmoid function
SIGMOFFSET =  0.0;          % Offset of the sigmoid function
SIGMOUT    =  300.0;        % Output gain of the sigmoid function

% Coefficient of the leaky integrator (if 0.0, integrator is removed)

INTEGRA    =  0.0;

%------------------------------------------------------------------------------
% Init delay lines
%
% 27.11.1995 Rami Hanninen
%
% Replaced the traditional ring-buffer delay line implementation with a more
% direct vector approach.  Even if the ring-buffered implementation is usually
% the most efficient method to implement a delay line, this is not the case
% with MatLab because executing for-loops and mod-operations in MatLab is much
% slower than the fast array operations MatLab offers.
%
% In the new implementation, the newest element in the delay line is LINE(1)
% and the oldest is LINE(LINELENGTH), exept with the lower delay line which is
% reversed so that the newest element is LOWER(LENGTH) and the oldest is
% LOWER(1).
%

JET   = zeros(1,JETLENGTH);   % Delay line for the air jet
UPPER = zeros(1,FLUTELENGTH); % Upper trail of the digital waveguide
LOWER = zeros(1,FLUTELENGTH); % Lower trail of the digital waveguide

%------------------------------------------------------------------------------
% Allocate memory for output vectors
%

UPPOUT = zeros(1,SAMPLES);
LOWOUT = zeros(1,SAMPLES);
JETOUT = zeros(1,SAMPLES);     
SIGOUT = zeros(1,SAMPLES);
VENOUT = zeros(NROFVENTS,SAMPLES);

%------------------------------------------------------------------------------
% Input of the flute model
%
% Use external random matrix if such is provided (handy when debugging)
%

if exist('PRESSURE') == 0
  PRESSURE  = NOISEGAIN * 2 *(rand(1,SAMPLES) - 0.5);  % Input to the model
end

INPUTAMPL = INPUTGAIN*[linspace(0,1,2000) ones(1,SAMPLES-4000) zeros(1,2000)];

%PRESSURE = [-1000 zeros(1,SAMPLES-1)];  % For measuring the impulse response
%INPUTGAIN = [1 zeros(1,SAMPLES-1)];

%------------------------------------------------------------------------------
% State variables
%

LOSSDELAY    = 0.0;  % Unit delay of the freq-dependent loss filter
REFLDELAYX   = 0.0;  % First unit delay of the reflection function
REFLDELAYY   = 0.0;  % Second unit delay of the reflection function
SIGMINPUT    = 0.0;  % Input of the sigmoid function
SIGMOUTPUT   = 0.0;  % Output of the sigmoid function
DCXOUTPUT0   = 0.0;  % output of the DC killer filter
DCXOUTPUT1   = 0.0;  % Last output of the DC killer filter
INTEGRINPUT  = 0.0;  % Input of the leaky integrator
INTEGROUTPUT = 0.0;  % Output of the integrator
VENTOUTPUT   = 0.0;  % 3-port vent output

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Flute model description by Vesa Valimaki:
%
% The computation of the flute model is described below.
% The parts of the waveguide flute algorithm are as follows
% (in this order):
%
%  - add input (noise signal) and signal from the integrator
%  - multiply them by the INPUTGAIN
%  - delay the above result using jet delay line
%  - feed the delayed signal into the sigmoid waveshaper
%  - add the output of the sigmoid function and the reflecting signal
%    from the lower delay line of the waveguide
%  - FRACTIONAL DELAY 3-PORT HERE !!!
%  - update the pointers to the delay lines (circular buffers) and move
%    them to the other end of the lines
%  - take a sample from the end of the upper trail and feed it to
%    the reflection filter
%  - run the reflection filter one step
%  - subtract the output of the reflection function from its input
%  - this result is the output of the model which has to be differentiated
%  - feed the output of the reflection filter to the lower delay line
%  - take a sample out of the other end (i.e., beginning) of the lower line
%  - add it to the output of the sigmoid function (multiplied by DIRFDBACK)
%  - multiply this result by VOICEDGAIN and feed it to the leaky integrator
%  - run one step of the integration (DC gain must be = 1)
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

for I = 1:SAMPLES

  % Print progress indicator

  if rem(I,100) == 0
    fprintf(1,'Sample %d (%.2f %%)\n',I,100*I/SAMPLES);
  end;

  % Open and close the vent. Note that we also adjust the jet line length.
  % When vent is closed, some extra zeros are added to the jet line. The
  % extra zeros are cut away next time the jet line advances.

  if I == OPENVENTHERE
    JETLENGTH  = JETLENGTH1;
    JET = JET(1:JETLENGTH);
    BRI = BR;
  elseif I == CLOSEVENTHERE
    JETLENGTH  = JETLENGTH2;
    JET = [ JET zeros(1,LENGTH) ]; % Too many zeros, will be cut away
    BRI = 0.0;
  end

  % Feed jet line

  JET(1) = INPUTAMPL(I) * (PRESSURE(I) + INTEGROUTPUT);

  % Sigmoid nonlinearity:

  SIGMOUTPUT = SIGMOUT*tanh(INPUTAMPL(I)*SIGMOFFSET-SIGMIN*JET(JETLENGTH));

  % DC killer (a 1st order high-pass filter, direct form II)

  TEMP       = SIGMOUTPUT + DCA * DCXOUTPUT1;
  DCXOUTPUT0 = TEMP - DCXOUTPUT1;  % differentiation removes DC component
  DCXOUTPUT1 = TEMP;

  % Add (subtract) the output of the sigmoid function (DC killed) and the
  % reflected signal from the end of the lower delay line and feed the
  % result into the beginning of the upper delay line.

  UPPER(1) = DCXOUTPUT0 + BACKGAIN * LOWER(1);

  % FRACTIONAL DELAY 3-PORT:
  % COEFFS  = Lagrange interpolator coefficients
  % DEILINE = delay line (state) of the deinterpolator
  % BR & AR = coefficients of the reflection filter

  %
  % Because our delay lines are now ordinary vectors, the 3-port computation
  % becomes a very simple matrix operation (the old implementation that used
  % ring buffered delay lines took over 20 lines of code and included two
  % slow for-loops).
  %
  % The input interpolation is computed like this:
  %
  % VENTINPUT = (UPPER + LOWER) * COEFF
  %
  % UPPER:      1 x LENGTH vector     Upper delay line
  % LOWER:      1 x LENGTH vector     Lower delay line (reversed)
  % COEFF:      LENGTH x VENTS matrix Lagrangian interpolation coefficients
  % VENTINPUT : 1 x VENTS vector      Input for N vents
  %
  % Note, that this implementation allows us to compute the input for multiple
  % 3-port vents in just a single operation.
  %
  % The output of the 3-port is computed like this
  %
  % VENTOUTPUT = BRI * VENTINPUT - AR * VENTOUTPUT
  %
  % BRI: scalar
  % AR:  scalar
  % VENTOUTPUT: 1 x VENTS vector
  %
  % And again, note this allows the simultanious computation of several vents.
  % Note also that VENTOUTPUT holds the previous vent outputs
  %
  % When combined, we finally get (ta-daa) the following:

  VENTOUTPUT  = BRI * ((UPPER + LOWER) * COEFFS) - AR * VENTOUTPUT;

  % Deinterpolation reverses the process

  DEINTER = VENTOUTPUT * COEFFS';
  UPPER   = UPPER + DEINTER;
  LOWER   = LOWER + DEINTER;

  % Boundary losses (1st order all-pole low-pass filter):

  LOSSDELAY = LOSSSCALE * LOSSGAIN * UPPER(FLUTELENGTH) + LOSSFREQ * LOSSDELAY;

  REFLDELAYY = RB0 * LOSSDELAY + RB1 * (REFLDELAYX - REFLDELAYY);
  REFLDELAYX = LOSSDELAY;  % Update unit delay of the reflection filter

  % Feed reflected signal into the beginning of the lower delay line

  LOWER(FLUTELENGTH) = REFLDELAYY;

  % Feedback loop:

  INTEGRINPUT = VOICEDGAIN * (LOWER(1) + DIRFDBACK * DCXOUTPUT0);

  % Integration:

  INTEGROUTPUT = (1 - INTEGRA) * INTEGRINPUT + INTEGRA * INTEGROUTPUT;

  % Output of the model:
  % Subtract output of refl. filter from its input (and differentiate):

  UPPOUT(1,I) = UPPER(FLUTELENGTH);
  LOWOUT(1,I) = LOWER(1);
  JETOUT(1,I) = JET(JETLENGTH);
  SIGOUT(1,I) = LOSSDELAY - REFLDELAYY;
  VENOUT(:,I) = VENTOUTPUT';

  %SIGOUT(1,I) = SIGMOUTPUT;
  %DCXOUT(1,I) = DCXOUTPUT0;

  % Move delay lines

  JET   = [ 0 JET(1:JETLENGTH-1)    ]; %  -->
  UPPER = [ 0 UPPER(1:FLUTELENGTH-1)]; %  -->
  LOWER = [ LOWER(2:FLUTELENGTH)   0]; % <--

end; % End of the main loop

fprintf(1,'\nDone. Execution time %.1f s\n\n',etime(clock,STARTTIME));
